package healthcheck

import (
	"context"
	"fmt"
	"net"
	"net/http"
	"net/http/httptest"
	"strconv"
	"strings"
	"testing"

	"github.com/open-io/oio-exporter/common"
	"github.com/open-io/oio-exporter/testutils"
)

func makeFakeHTTPService(route, data string, statusCode int) (*httptest.Server, string, int) {
	server := httptest.NewServer(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		if r.URL.Path != route {
			return
		}

		w.WriteHeader(statusCode)
		fmt.Fprint(w, data)
	}))

	parts := strings.Split(server.URL, ":")
	ip := parts[1][2:]
	port, err := strconv.ParseInt(parts[2], 10, 64)
	if err != nil {
		log.Fatalln(err)
	}

	return server, ip, int(port)
}

func runFakeTCPService(l net.Listener, data string) {
	for {
		conn, err := l.Accept()
		if err != nil {
			// https://grokbase.com/t/gg/golang-nuts/131k81bse0/go-nuts-use-of-closed-network-connection-error-from-tcpconn-read
			netOpError, ok := err.(*net.OpError)
			if ok && netOpError.Err.Error() == "use of closed network connection" {
				return
			}

			log.Warn(err)
			return
		}
		go func(conn net.Conn) {
			defer conn.Close()
			var res []byte
			_, err := conn.Read(res)
			if err != nil {
				log.Fatalln(err)
			}
			_, err = conn.Write([]byte(data))
			if err != nil {
				log.Fatalln(err)
			}
		}(conn)
	}
}

func makeFakeTCPService(data string) (net.Listener, int) {
	listener, err := net.Listen("tcp", "127.0.0.1:0")
	if err != nil {
		log.Fatalln(err)
	}
	go runFakeTCPService(listener, data)

	addr := listener.Addr().String()

	port, err := strconv.ParseInt(strings.Split(addr, ":")[1], 10, 64)
	if err != nil {
		log.Fatalln(err)
	}

	return listener, int(port)
}

func makeFakeServices(healthchecks map[string]healthcheck) ([]*common.Service, map[string]float64, []*httptest.Server, []net.Listener) {
	expected := map[string]float64{}
	targets := make([]*common.Service, 0, 5)
	httpServers := make([]*httptest.Server, 0, 5)
	tcpListeners := make([]net.Listener, 0, 5)

	httpServer, ip, port := makeFakeHTTPService(healthchecks["account"].(*HTTPCheck).Route, "", 200)
	targets = append(targets, &common.Service{
		ID:     "OPENIO-account-0",
		IP:     ip,
		Port:   port,
		Volume: "",
		Type:   "account",
		Conf:   map[string]string{},
	})
	expected["OPENIO-account-0"] = float64(1)
	httpServers = append(httpServers, httpServer)

	tcpListener, port := makeFakeTCPService("OK \n")
	targets = append(targets, &common.Service{
		ID:     "OPENIO-meta0-0",
		IP:     "127.0.0.1",
		Port:   port,
		Volume: "",
		Type:   "meta0",
		Conf:   map[string]string{},
	})
	expected["OPENIO-meta0-0"] = float64(1)
	tcpListeners = append(tcpListeners, tcpListener)

	tcpListener, port = makeFakeTCPService("OK \n")
	targets = append(targets, &common.Service{
		ID:     "OPENIO-beanstalkd-0",
		IP:     "127.0.0.1",
		Port:   port,
		Volume: "",
		Type:   "beanstalkd",
		Conf:   map[string]string{},
	})
	expected["OPENIO-beanstalkd-0"] = float64(1)
	tcpListeners = append(tcpListeners, tcpListener)

	tcpListener, port = makeFakeTCPService("KO \n")
	targets = append(targets, &common.Service{
		ID:     "OPENIO-beanstalkd-1",
		IP:     "127.0.0.1",
		Port:   port,
		Volume: "",
		Type:   "beanstalkd",
		Conf:   map[string]string{},
	})
	expected["OPENIO-beanstalkd-1"] = float64(0)
	tcpListeners = append(tcpListeners, tcpListener)

	httpServer, ip, port = makeFakeHTTPService(healthchecks["keystone"].(*HTTPCheck).Route, "", 401)
	targets = append(targets, &common.Service{
		ID:     "OPENIO-keystone-0",
		IP:     ip,
		Port:   port,
		Volume: "",
		Type:   "keystone",
		Conf:   map[string]string{},
	})
	expected["OPENIO-keystone-0"] = float64(1)
	httpServers = append(httpServers, httpServer)

	return targets, expected, httpServers, tcpListeners
}

func TestHealtcheckModule(t *testing.T) {
	t.Parallel()

	healthchecksPath := "../healthchecks.yml"

	m := common.MakeMetricData()
	d := MakeDaemon("", healthchecksPath)

	healthchecks, err := d.parseHealthchecks(healthchecksPath, "")
	if err != nil {
		t.Fatalf("Error parsing healthchecks: %s", err)
	}

	targets, expected, httpServers, tcpListeners := makeFakeServices(healthchecks)
	for _, httpServer := range httpServers {
		defer httpServer.Close()
	}
	for _, tcpListener := range tcpListeners {
		defer tcpListener.Close()
	}

	if err := d.Init(nil, targets); err != nil {
		t.Fatalf("Init error: %s", err)
	}

	if err := d.Collect(m); err != nil {
		t.Fatalf("Collect error: %s", err)
	}

	if err := d.Close(context.TODO()); err != nil {
		t.Fatalf("Close error: %s", err)
	}

	mdt := testutils.NewMetricDataTest()
	for service, expectedValue := range expected {
		mdt.ForMetricWithLabels(
			testutils.NewMetricAssertion().Equals(expectedValue),
			"probe_success",
			"service", service)
	}
	mdt.Check(t, m)
}
