package redis

import (
	"context"
	"encoding/json"
	"regexp"
	"strconv"
	"strings"
	"sync"

	"github.com/go-redis/redis"
	"github.com/open-io/oio-exporter/common"
	_log "github.com/open-io/oio-exporter/common/log"
	"github.com/open-io/oio-exporter/daemon"
)

var log = _log.New("redis")

var scriptGetAccounts = redis.NewScript(`
    return redis.call("hgetall", "accounts:")
`)

var scriptGetContCount = redis.NewScript(`
    local key = "containers:" .. KEYS[1]
    return redis.call("ZCOUNT", key, 0, 0)
`)

var scriptAcctInfo = redis.NewScript(`
	local accts = redis.call("hgetall", "accounts:")
	local res = {}
	local index = 0
	for i, acc in ipairs(accts) do
        if i % 2 == 1 then
                local acct_key = "account:" .. acc;
                res[index] = {acc, redis.call('HGET', acct_key, 'bytes'), redis.call("HGET", acct_key, 'objects')}
                index = index + 1
        end;
	end;
	return cjson.encode(res);
`)

type client struct {
	Client     *redis.Client
	IsAccount  bool
	IsCH       bool
	IsBucketDB bool
	Service    *common.Service
}

type redisDaemon struct {
	daemon.Daemon

	clients   []*client
	namespace string
}

var dbRegex = regexp.MustCompile(`^db(\d+):keys=(\d+),expires=(\d+)`)

func MakeDaemon(namespace string) *redisDaemon {
	return &redisDaemon{
		clients:   []*client{},
		namespace: namespace,
	}
}

func (d *redisDaemon) ID() string {
	return "Redis"
}

func (d *redisDaemon) Init(services common.ServicesMap, targets []*common.Service) error {
	d.reload(services)

	return nil
}

func (d *redisDaemon) OnServicesChange(services common.ServicesMap) error {
	log.Noticeln("redis module", "reloading config")

	d.reload(services)

	return nil
}

func (d *redisDaemon) reload(services common.ServicesMap) {
	d.cleanup()

	redisInstances, ok := services["redis"]
	if !ok || len(redisInstances) == 0 {
		return
	}

	redissentinelInstances, ok := services["redissentinel"]
	if ok && len(redissentinelInstances) > 0 {
		redisInstances = append(redisInstances, redissentinelInstances...)
	}

	clients := []*client{}
	var wg sync.WaitGroup
	var mutex sync.Mutex
	for _, svc := range redisInstances {
		wg.Add(1)
		go func(service *common.Service) {
			defer wg.Done()

			newClient := &client{
				Service: service,
				Client: redis.NewClient(&redis.Options{
					Addr: service.IP + ":" + strconv.Itoa(service.Port),
				}),
			}
			val, err := newClient.Client.HLen("accounts:").Result()
			// TODO: handle error
			if err == nil && val > 0 {
				newClient.IsAccount = true
			}

			mutex.Lock()
			log.Noticeln("new redis client:", newClient.Client)
			clients = append(clients, newClient)
			mutex.Unlock()
		}(svc)
	}

	wg.Wait()

	d.clients = clients
}

func (d *redisDaemon) Collect(m *common.MetricData) error {
	var wg sync.WaitGroup
	for _, c := range d.clients {
		if c.IsAccount {
			wg.Add(1)
			go func(c *client) {
				defer wg.Done()

				d.collectAccount(c, m)
			}(c)
		}
		wg.Add(1)
		go func(c *client) {
			defer wg.Done()

			d.collectStats(c, m)
		}(c)
	}

	wg.Wait()

	return nil
}

func (d *redisDaemon) cleanup() {
	for _, c := range d.clients {
		if err := c.Client.Close(); err != nil {
			log.Warnf("failed to close redis client: %s", c.Client)
		}
	}
	d.clients = []*client{}
}

func (d *redisDaemon) Close(ctx context.Context) error {
	d.cleanup()

	return nil
}

func (d *redisDaemon) sendStat(metric string, value float64, labels [][]string, m *common.MetricData) {
	m.Set(&common.Metric{
		Name:   "redis_" + metric,
		Labels: labels,
		Value:  value,
	})
}

func (d *redisDaemon) collectStats(c *client, m *common.MetricData) {
	data, err := c.Client.Info().Result()
	if err != nil {
		m.IncrError("redis", "redis")
		log.Warnln("error while collecting redis stats", c, err)
		return
	}
	for _, line := range strings.Split(data, "\r\n") {
		kv := strings.SplitN(line, ":", 2)
		if len(kv) != 2 {
			continue
		}

		stat := kv[0]
		valStr := kv[1]

		labels := [][]string{
			[]string{"namespace", d.namespace},
			[]string{"service", c.Service.ID},
			[]string{"service_type", c.Service.Type},
			[]string{"service_addr", c.Client.Options().Addr},
		}

		if matches := dbRegex.FindStringSubmatch(line); len(matches) == 4 {
			valKeys, errKeys := strconv.ParseFloat(matches[2], 64)
			valExpires, errExpires := strconv.ParseFloat(matches[3], 64)
			if errKeys != nil || errExpires != nil {
				// TODO log
				continue
			}
			labels = append(labels, []string{"db", matches[1]})

			labelsKeys := make([][]string, len(labels))
			copy(labelsKeys, labels)
			d.sendStat("db_keys", valKeys, append(labelsKeys, []string{"type", "keys"}), m)

			labelsExpires := make([][]string, len(labels))
			copy(labelsExpires, labels)
			d.sendStat("db_keys", valExpires, append(labelsExpires, []string{"type", "expires"}), m)
			continue
		}

		val, errFloat := strconv.ParseFloat(valStr, 64)

		switch stat {
		case "redis_mode":
			stat = "is_" + valStr
			val = 1
			errFloat = nil

		case "uptime_in_seconds":
			stat = "uptime_seconds"

		case "connected_clients":
			stat = "clients"
			labels = append(labels, []string{"type", "connected"})
		case "blocked_clients":
			stat = "clients"
			labels = append(labels, []string{"type", "connected"})

		case "used_memory":
			stat = "memory_bytes"
			labels = append(labels, []string{"type", "total"})
		case "used_memory_rss":
			stat = "memory_bytes"
			labels = append(labels, []string{"type", "rss"})
		case "used_memory_overhead":
			stat = "memory_bytes"
			labels = append(labels, []string{"type", "overhead"})
		case "used_memory_dataset":
			stat = "memory_bytes"
			labels = append(labels, []string{"type", "dataset"})
		case "used_memory_lua":
			stat = "memory_bytes"
			labels = append(labels, []string{"type", "lua"})
		case "used_memory_scripts":
			stat = "memory_bytes"
			labels = append(labels, []string{"type", "scripts"})
		case "used_memory_peak":
			stat = "used_memory_bytes_max"

		case "mem_fragmentation_ratio":
			stat = "memory_fragmentation_ratio"

		case "active_defrag_running":
			stat = "active_defrag_is_running"
		case "active_defrag_hits":
			stat = "active_defrag"
			labels = append(labels, []string{"type", "hits"})
		case "active_defrag_misses":
			stat = "active_defrag"
			labels = append(labels, []string{"type", "misses"})
		case "active_defrag_key_hits":
			stat = "active_defrag"
			labels = append(labels, []string{"type", "key_hits"})
		case "active_defrag_key_misses":
			stat = "active_defrag"
			labels = append(labels, []string{"type", "key_misses"})

		case "lazyfree_pending_objects":

		case "rdb_changes_since_last_save":
		case "rdb_bgsave_in_progress":
		case "rdb_last_save_time":
			stat += "_seconds"
		case "rdb_last_bgsave_status":
			if valStr == "ok" {
				val = 1.0
			} else {
				val = 0.0
			}
			errFloat = nil
		case "rdb_last_bgsave_time_sec":
			stat = "rdb_last_bgsave_time_seconds"
		case "rdb_current_bgsave_time_sec":
			stat = "rdb_current_bgsave_time_seconds"
		case "rdb_last_cow_size":
			stat += "_bytes"

		case "aof_enabled":
		case "aof_rewrite_in_progress":
		case "aof_rewrite_scheduled":
		case "aof_last_rewrite_time_sec":
			stat = "aof_last_rewrite_time_seconds"
		case "aof_current_rewrite_time_sec":
			stat = "aof_current_rewrite_time_seconds"
		case "aof_last_bgrewrite_status":
			if valStr == "ok" {
				val = 1.0
			} else {
				val = 0.0
			}
			errFloat = nil
		case "aof_last_write_status":
			if valStr == "ok" {
				val = 1.0
			} else {
				val = 0.0
			}
			errFloat = nil
		case "aof_last_cow_size":
			stat += "_bytes"
		case "aof_current_size":
			stat += "_bytes"
		case "aof_base_size":
			stat += "_bytes"
		case "aof_pending_rewrite":
		case "aof_buffer_length":
		case "aof_rewrite_buffer_length":
		case "aof_pending_bio_fsync":
		case "aof_delayed_fsync":
			stat += "_total"

		case "loading":
		case "loading_start_time":
			stat += "_seconds"
		case "loading_total_bytes":
		case "loading_loaded_bytes":
		case "loading_eta_seconds":

		case "rejected_connections":
			stat = "rejected_connections_total"
		case "total_connections_received":
			stat = "connections_total"
		case "total_commands_processed":
			stat = "commands_total"
		case "instantaneous_ops_per_sec":
			stat = "commands_ops"

		case "total_net_input_bytes":
			stat = "io_bytes_total"
			labels = append(labels, []string{"direction", "in"})
		case "total_net_output_bytes":
			stat = "io_bytes_total"
			labels = append(labels, []string{"direction", "out"})
		case "instantaneous_input_kbps":
			stat = "io_kbps"
			labels = append(labels, []string{"direction", "in"})
		case "instantaneous_output_kbps":
			stat = "io_kbps"
			labels = append(labels, []string{"direction", "out"})

		case "sync_full":
		case "sync_partial_ok":
			stat = "sync_partial"
			labels = append(labels, []string{"status", "ok"})
		case "sync_partial_err":
			stat = "sync_partial"
			labels = append(labels, []string{"status", "err"})

		case "expired_keys":
			stat = "keys_total"
			labels = append(labels, []string{"type", "expired"})
		case "evicted_keys":
			stat = "keys_total"
			labels = append(labels, []string{"type", "evicted"})

		case "keyspace_hits":
			stat = "keyspace_total"
			labels = append(labels, []string{"type", "hits"})
		case "keyspace_misses":
			stat = "keyspace_total"
			labels = append(labels, []string{"type", "misses"})

		case "pubsub_channels":
		case "pubsub_patterns":
		case "latest_fork_usec":
			stat = "latest_fork_duration_seconds"
			val /= 1e6
		case "migrate_cached_sockets":
		case "slave_expires_tracked_keys":
		case "role":
			labels = append(labels, []string{"role", valStr})
			val = 1
			errFloat = nil
		case "repl_backlog_active":
		case "repl_backlog_size":
			stat += "_bytes"
		case "master_link_status":
			labels = append(labels, []string{"status", valStr})
			val = 1
			errFloat = nil
		case "master_last_io_seconds_ago":
			stat = "master_last_io_seconds"
		case "master_sync_in_progress":
		case "slave_priority":
		case "slave_read_only":
		case "master_sync_left_bytes":
		case "master_sync_last_io_seconds_ago":
			stat = "master_sync_last_io_seconds"
		case "master_link_down_since_seconds":
		case "connected_slaves":
		case "min_slaves_good_slaves":

		default:
			/* ignore unknown commands */
			continue
		}

		if errFloat != nil {
			// TODO log
			continue
		}
		d.sendStat(stat, val, labels, m)
	}
}

func (d *redisDaemon) collectAccount(c *client, m *common.MetricData) {
	acctInfo, err := scriptAcctInfo.Run(c.Client, []string{}, 0).Result()
	if err != nil {
		log.Warnln("error while collecting account info", c, err)
		return
	}
	acctObj := map[string][]string{}
	err = json.Unmarshal([]byte(acctInfo.(string)), &acctObj)
	if err != nil {
		m.IncrError("redis", "account")
		log.Warnln("error while collecting account info", c, err)
		return
	}
	for _, data := range acctObj {
		labels := [][]string{
			[]string{"namespace", d.namespace},
			[]string{"service", c.Service.ID},
			[]string{"service_type", c.Service.Type},
			[]string{"service_addr", c.Client.Options().Addr},
		}
		valSize, err := strconv.ParseFloat(data[1], 64)
		if err != nil {
			m.IncrError("redis", "account")
			log.Warnln("error while collecting account objects", c, err)
			continue
		}
		valObj, err := strconv.ParseFloat(data[2], 64)
		if err != nil {
			m.IncrError("redis", "account")
			log.Warnln("error while collecting account objects", c, err)
			continue
		}
		labels = append(labels, []string{"account", data[0]})
		m.Set(&common.Metric{
			Name:   "redis_account_size_bytes",
			Labels: labels,
			Value:  valSize,
		})
		m.Set(&common.Metric{
			Name:   "redis_account_objects",
			Labels: labels,
			Value:  valObj,
		})
	}

	accounts, err := scriptGetAccounts.Run(c.Client, []string{}, 0).Result()
	if err != nil {
		m.IncrError("redis", "account")
		log.Warnln("error while listing accounts", c, err)
		return
	}
	for _, acct := range accounts.([]interface{}) {
		if acct == "1" {
			continue
		}
		count, err := scriptGetContCount.Run(c.Client, []string{acct.(string)}, 1).Result()
		if err != nil {
			m.IncrError("redis", "account")
			log.Warnln("error while collecting account containers", acct, err)
			continue
		}
		m.Set(&common.Metric{
			Name: "redis_account_containers",
			Labels: [][]string{
				[]string{"namespace", d.namespace},
				[]string{"service", c.Service.ID},
				[]string{"service_type", c.Service.Type},
				[]string{"service_addr", c.Client.Options().Addr},
				[]string{"account", acct.(string)},
			},
			Value: float64(count.(int64)),
		})
	}
}
